
(*
A verified abstract machine forfunctional coroutines - Coq formalization(companion technical report)

by Tristan Crolard

CNAM Paris, France

Email: tristan.crolard@cnam.fr




*)
(*
Abstract
Functional coroutines are a restricted form of control mechanism, where each continuation comes with its local environment. This restriction was originally obtained by considering a constructive version of Parigot's classical natural deduction which is sound and complete for the Constant Domain logic. In this article, we present a refinement of de Groote's abstract machine which is proved to be correct for functional coroutines. Therefore, this abstract machine also provides a direct computational interpretation of the Constant Domain logic.

This companion technical report contains the full Coq formalization (all the proofs are available in the Coq source file). The section numbering is the same as in the article for easy cross referencing. 
*)
(*
1 Introduction 


2 Dependency relations 


3 Abstract machines 
*)
Require Import Utf8.
Add LoadPath "~/Sources/Coq/Tactics".
Require Import CpdtTactics LibTactics TwelfTactics.

Require Import Arith List String.
Require Import Coq.extraction.ExtrOcamlString.

Definition ble_nat (n : nat) (m : nat) : bool := beq_nat (n - m) 0.
Notation "n == m" := (beq_nat n m) (at level 70, no associativity).
Notation "n ≦ m" := (ble_nat n m) (at level 70, no associativity).

Reserved Notation "m_1  _-_  m_2  _=_  m" (at level 30).

Inductive minus : nat → nat → nat → Prop :=

| Minus_1 : ∀ m_1,
m_1  _-_  0  _=_  m_1 

| Minus_2 : ∀ m_1 m_2 m,
(m_1  _-_  m_2  _=_  m)/(S(m_1)  _-_  S(m_2)  _=_  m) 

where "m_1  _-_  m_2  _=_  m" := (minus m_1 m_2 m).

Fixpoint f (m : nat) : (m  _-_  m  _=_  0) :=
  match m with
  | 0 => Minus_1 0
  | S(n) => Minus_2 n n 0 (f n)
  end.

Lemma minus_trivial : ∀ n, not (0=S n).
Proof.
  auto.
Qed.

Lemma minus_id : ∀ m, m  _-_  m  _=_  0.
Proof.
  exact f.
Qed.

Hint Resolve minus_id.

Lemma minus_unique : ∀ m_2 m_1 m m', m_1  _-_ m_2  _=_  m → m_1  _-_  m_2  _=_  m' → m=m'.
Proof.
  intros.
  induction H.
  - inversion H0.
    auto.
  - inversion H0.
    auto.
Qed.

Hint Resolve minus_unique : gen_subst_db.

Lemma minus_succ : ∀ n_3 n_2 n_1, n_1  _-_ (S n_2)  _=_  n_3 → n_1  _-_  n_2  _=_  (S n_3).
Proof.
  intros n_3 n_2.
  induction n_2.
  - intros n_1 𝒟. 
    inversion 𝒟.
    inversion H1.
    apply Minus_1.
  - intros n_1 𝒟. 
    inversion 𝒟.
    apply Minus_2.
    apply IHn_2.
    exact H1.
Qed.

Lemma minus_swap : ∀ n_3 n_2 n_1, n_1  _-_  n_2  _=_  n_3 → n_1  _-_  n_3  _=_  n_2.
Proof.
  intros n_3 n_2. 
  induction n_2. 
  - intros n_1 𝒟. 
    inversion 𝒟. 
    apply minus_id. 
  - intros n_1 𝒟. 
    inversion 𝒟. 
    apply minus_succ. 
    apply Minus_2. 
    apply IHn_2. 
    exact H1.
Qed.


Definition vector := list nat.
Definition table := list vector.

Reserved Notation "ℐ  _( n )_   _=_  m" (at level 30).

Inductive fetch {A : Type} : list A → nat → A → Prop :=

| Fetch_1 : ∀ ℐ (n:A),
 (n::ℐ) _(0)_  _=_  n 

| Fetch_2 : ∀ ℐ (n:A) n_1 n_2,
(ℐ _(n_1)_  _=_  n_2)/((n::ℐ) _(S n_1)_  _=_  n_2) 

where "ℐ  _( n )_   _=_  m" := (@fetch _ ℐ n m).


Lemma fetch_unique : ∀ A n ℐ (y:A) (y':A), ℐ _(n)_  _=_ y → ℐ _(n)_  _=_  y' → y=y'.
Proof.
  introv. revert_last.
  intros y1 H H0. revert H0. revert y1.
  induction H; 
    (intros y2 D2; inversion D2; repeat progress ( eauto || econstructor || gen_subst || crush )).
Qed.

Hint Resolve fetch_unique : gen_subst_db.

Reserved Notation "n  _-_  ℐ  _(_  l  _)_   _=_  g" (at level 30).

Inductive compute : list nat → nat → nat → nat → Prop :=

| Compute_1 : ∀ ℐ n l g k,
(ℐ _(l)_  _=_  k) → (n _-_ k  _=_  g)/(n  _-_  ℐ _(_ l _)_  _=_  g) 

where "n  _-_  ℐ  _(_  l  _)_   _=_  g" := (compute ℐ n l g).

Lemma compute_unique : ∀ ℐ n l y y', n  _-_  ℐ _(_ l _)_  _=_  y → n  _-_  ℐ _(_ l _)_  _=_  y' → y=y'.
Proof.
  introv. revert_last.
  intros y1 H H0. revert H0. revert y1.
  induction H; 
    (intros y2 D2; inversion D2; repeat progress ( eauto || econstructor || gen_subst || crush )).
Qed.

Hint Resolve compute_unique : gen_subst_db.

Reserved Notation "n  _∈_  ℐ" (at level 30).

Inductive member : list nat → nat → Prop :=

| Member_1 : ∀ ℐ n,
n _∈_ (n::ℐ) 

| Member_2 : ∀ ℐ n n',
(n _∈_ ℐ)/(n _∈_ (n'::ℐ)) 

where "n  _∈_  ℐ" := (member ℐ n).

Lemma domain : ∀ ℐ k, k _∈_ ℐ→∃n,ℐ _(n)_  _=_ k.
Proof.
intros ℐ k D.
induction D.
- exists 0.
  apply Fetch_1.
- destruct IHD.
  exists (S x).
  apply Fetch_2.
  exact H.
Qed.

(*3.1 Safe λ_ct-terms *)
Inductive term : Type :=
| Var (n : nat)
| App (t_1 : term) (t_2 : term)
| Lambda (t : term)
| Catch (t : term)
| Throw (α : nat) (t : term).

Notation "'‘' x '‘'" := (Var x) (at level 80).
Notation "t_1 '_*_' t_2" := (App t_1 t_2) (at level 80).
Notation "'_λ_' t" := (Lambda t) (at level 80).

Notation "'catch'" := Catch (at level 80).
Notation "'throw'" := Throw (at level 80).
Notation "'get-context'" := Catch (at level 80).
Notation "'set-context'" := Throw (at level 80).

Reserved Notation  "'Safe_' '(' n ')^(' ℐ , ℐ_μ ')' ( t )" (at level 30).

Inductive safe : term → list nat → list (list nat) → nat → Prop :=

| Safe_1 : ∀ ℐ ℐ_μ n g k,
(n _-_ g  _=_  k) → (k _∈_ ℐ)/(Safe_( n)^( ℐ,ℐ_μ)(‘g‘)) 

| Safe_2 : ∀ ℐ ℐ_μ n t u,
(Safe_( n)^( ℐ,ℐ_μ)(t)) → (Safe_( n)^( ℐ,ℐ_μ)(u))/(Safe_( n)^( ℐ,ℐ_μ)(t _*_ u)) 

| Safe_3 : ∀ ℐ ℐ_μ n t,
(Safe_( S n)^( (S n::ℐ),ℐ_μ)(t))/(Safe_( n)^( ℐ,ℐ_μ)( _λ_ t)) 

| Safe_4 : ∀ ℐ ℐ_μ n t,
(Safe_( n)^( ℐ,(ℐ::ℐ_μ))(t))/(Safe_( n)^( ℐ,ℐ_μ)(catch t)) 

| Safe_5 : ∀ ℐ ℐ' ℐ_μ n t α,
(ℐ_μ _(α)_   _=_  ℐ') → (Safe_( n)^( ℐ',ℐ_μ)(t))/(Safe_( n)^( ℐ,ℐ_μ)(throw α t)) 

where "'Safe_' '(' n ')^(' ℐ , ℐ_μ ')' ( t )" := (safe t ℐ ℐ_μ n).

(*3.2 From local indices to global indices *)
Reserved Notation "'↓_' '(' n ')^(' ℐ , ℐμ ')' ( t_1 ) = t_2" (at level 30).

Inductive litogi : term → list nat → list (list nat) → nat → term → Prop :=

| Litogi_1 : ∀ ℐ ℐ_μ n g l,
(n  _-_  ℐ _(_ l _)_  _=_  g)/(↓_( n)^( ℐ,ℐ_μ)(‘l‘)=(‘g‘)) 

| Litogi_2 : ∀ ℐ ℐ_μ n t u t' u',
(↓_( n)^( ℐ,ℐ_μ)(t)=t') → (↓_( n)^( ℐ,ℐ_μ)(u)=u')/(↓_( n)^( ℐ,ℐ_μ)(t _*_ u)=(t' _*_ u')) 

| Litogi_3 : ∀ ℐ ℐ_μ n t t',
(↓_( S n)^( (S n::ℐ),ℐ_μ)(t)=t')/(↓_( n)^( ℐ,ℐ_μ)( _λ_ t)=( _λ_ t')) 

| Litogi_4 : ∀ ℐ ℐ_μ n t t',
(↓_( n)^( ℐ,(ℐ::ℐ_μ))(t)=t')/(↓_( n)^( ℐ,ℐ_μ)(get-context t)=(catch t')) 

| Litogi_5 : ∀ ℐ ℐ' ℐ_μ n t t' α,
(ℐ_μ _(α)_   _=_  ℐ') → (↓_( n)^( ℐ',ℐ_μ)(t)=t')/(↓_( n)^( ℐ,ℐ_μ)(set-context α t)=(throw α t')) 


where "'↓_' '(' n ')^(' ℐ , ℐ_μ ')' ( t_1 ) = t_2" := (litogi t_1 ℐ ℐ_μ n t_2).


Lemma litogi_unique : ∀ ℐ ℐ_μ n x y y', ↓_( n)^( ℐ,ℐ_μ)(x)=y → ↓_( n)^( ℐ,ℐ_μ)(x)=y' → y=y'.
Proof.
  introv. revert_last.
  intros y1 H H0. revert H0. revert y1.
  induction H; 
    (intros y2 D2; inversion D2; repeat progress ( eauto || econstructor || gen_subst || crush )).
Qed.

Hint Resolve litogi_unique : gen_subst_db.

Lemma safe_image :  ∀ ℐ ℐ_μ n t', Safe_( n)^( ℐ,ℐ_μ)(t')→∃t,↓_( n)^( ℐ,ℐ_μ)(t)=t'.
Proof.
  intros ℐ ℐ_μ n t D.
  induction D.
  - apply domain in H0.
    apply minus_swap in H.
    destruct H0 as [ l D111 ].
    exists (‘l‘).
    apply Litogi_1.
    eapply Compute_1.
    exact D111.
    exact H.
  - destruct IHD1 as [ t' D11 ].
    destruct IHD2 as [ u' D12 ].
    exists (t' _*_ u').
    apply Litogi_2.
    exact D11.
    exact D12.
  - destruct IHD as [ t' D11 ].
    exists (_λ_ t').
    apply Litogi_3.
    exact D11.
  - destruct IHD as [ t' D11 ].
    exists (get-context t').
    apply Litogi_4.
    exact D11.
  - destruct IHD as [ t' D11 ].
    exists (set-context α t').
    eapply Litogi_5.
    exact H.
    exact D11.
Qed.

(*3.3 Abstract machine for λ_ct-terms *)
(*3.3.1 Closure, environment, stack and state *)
Inductive clos :=
| Cl (t : term) (ℰ : list clos) (ℰ_μ : list (list clos)).

Definition stack := list clos.
Definition c_env := list clos.
Definition k_env := list stack.

Inductive state :=
| St (t : term) (ℰ : c_env) (ℰ_μ : k_env) (𝒮 : stack).

Notation "⟨ t , ℰ , ℰ_μ , 𝒮 ⟩" := (St t ℰ ℰ_μ 𝒮) (at level 80).
Notation "[ t , ℰ , ℰ_μ ]" := (Cl t ℰ ℰ_μ) (at level 80).
(*3.3.2 Evaluation rules *)
Reserved Notation "σ_1 ↝ σ_2" (at level 30).

Inductive step : state → state → Prop :=

| K_var : ∀ k t ℰ ℰ' ℰ_μ ℰ_μ' 𝒮,
(ℰ _(k)_   _=_  [t,ℰ',ℰ_μ'])/(⟨ ‘k‘ , ℰ,ℰ_μ,𝒮⟩ ↝ ⟨t,ℰ',ℰ_μ',𝒮⟩) 

| K_app : ∀ t u ℰ ℰ_μ 𝒮,
⟨(t _*_ u), ℰ,ℰ_μ,𝒮⟩ ↝ ⟨t,ℰ,ℰ_μ,[u,ℰ,ℰ_μ]::𝒮⟩ 

| K_abs : ∀ t c ℰ ℰ_μ 𝒮,
⟨ _λ_ t, ℰ,ℰ_μ,c ::𝒮⟩ ↝ ⟨t,(c ::ℰ),ℰ_μ,𝒮⟩ 

| K_catch : ∀ t ℰ ℰ_μ 𝒮,
⟨catch t, ℰ,ℰ_μ,𝒮⟩ ↝ ⟨t,ℰ,(𝒮 ::ℰ_μ),𝒮⟩ 

| K_throw : ∀ t α ℰ ℰ_μ 𝒮 𝒮',
(ℰ_μ _(α)_   _=_  𝒮')/(⟨throw α t, ℰ,ℰ_μ,𝒮⟩ ↝ ⟨t,ℰ,ℰ_μ,𝒮'⟩) 

where "σ_1 ↝ σ_2" := (step σ_1 σ_2).

Lemma step_unique : ∀ σ_1 σ_2 σ_2', σ_1 ↝ σ_2 → σ_1 ↝ σ_2' → σ_2=σ_2'.
Proof.
  introv. revert_last.
  intros y1 H H0. revert H0. revert y1.
  induction H; 
    (intros y2 D2; inversion D2; repeat progress ( eauto || econstructor || gen_subst || crush )).
Qed.

Hint Resolve step_unique : gen_subst_db.

(*3.4 Abstract machine for safe λ_ct-terms (with local environments) *)
(*3.4.1 Closure, environment, stack and state *)
Inductive clos_l :=
| Cl_l (t : term) (ℒ : list clos_l) (ℒ_μ : list (list clos_l)) (ℰ_μ : list (list clos_l)).

Definition stack_l := list clos_l.
Definition l_env_l := list clos_l.
Definition m_env_l := list l_env_l.
Definition k_env_l := list stack_l.

Inductive state_l :=
| St_l (t : term) (ℒ : l_env_l) (ℒ_μ : m_env_l) (ℰ_μ : k_env_l) (𝒮 : stack_l).

Notation "⟨ t , ℒ , ℒ_μ , ℰ_μ , 𝒮 ⟩" := (St_l t ℒ ℒ_μ ℰ_μ 𝒮) (at level 80).
Notation "[ t , ℒ , ℒ_μ , ℰ_μ ]" := (Cl_l t ℒ ℒ_μ ℰ_μ) (at level 80).
(*3.4.2 Evaluation rules *)
Reserved Notation "σ_1 ↝ '^l' σ_2" (at level 30).

Inductive step_l : state_l → state_l → Prop :=

| L_var : ∀ k t ℒ ℒ' ℒ_μ ℒ_μ' ℰ_μ ℰ_μ' 𝒮,
(ℒ _(k)_   _=_  [t,ℒ',ℒ_μ',ℰ_μ'])/(⟨ ‘k‘ ,ℒ,ℒ_μ,ℰ_μ,𝒮⟩ ↝^l⟨ t, ℒ',ℒ_μ',ℰ_μ',𝒮⟩) 

| L_app : ∀ t u ℒ ℒ_μ ℰ_μ 𝒮,
⟨(t _*_ u),ℒ,ℒ_μ,ℰ_μ,𝒮⟩ ↝^l⟨ t,ℒ,ℒ_μ,ℰ_μ,[u,ℒ,ℒ_μ,ℰ_μ]::𝒮⟩ 

| L_abs : ∀ t c ℒ ℒ_μ ℰ_μ S',
⟨ _λ_ t, ℒ,ℒ_μ,ℰ_μ,c ::S'⟩ ↝^l⟨ t, (c :: ℒ),ℒ_μ,ℰ_μ,S'⟩ 

| L_catch : ∀ t ℒ ℒ_μ ℰ_μ 𝒮,
⟨get-context t, ℒ,ℒ_μ,ℰ_μ,𝒮⟩ ↝^l⟨ t,  ℒ,(ℒ ::ℒ_μ),(𝒮 ::ℰ_μ),𝒮⟩ 

| L_throw : ∀ t α ℒ ℒ' ℒ_μ ℰ_μ 𝒮 𝒮',
(ℒ_μ _(α)_   _=_  ℒ') → (ℰ_μ _(α)_   _=_  𝒮')/(⟨set-context α t,ℒ,ℒ_μ,ℰ_μ,𝒮⟩ ↝^l⟨ t,  ℒ',ℒ_μ,ℰ_μ,𝒮'⟩) 

where "σ_1 ↝ '^l' σ_2" := (step_l σ_1 σ_2).

Lemma step_l_unique : ∀ σ_1 σ_2 σ_2', σ_1 ↝^l σ_2 → σ_1 ↝^l σ_2' → σ_2=σ_2'.
Proof.
  introv. revert_last.
  intros y1 H H0. revert H0. revert y1.
  induction H; 
    (intros y2 D2; inversion D2; repeat progress ( eauto || econstructor || gen_subst || crush )).
Qed.

Hint Resolve step_l_unique : gen_subst_db.


(*
4 Bisimulations *)
(*4.1 Abstract machine for safe λ_ct-terms (with indirection tables) *)
(*4.1.1 Closure, environment, stack and state *)
Inductive clos_i :=
| Cl_i (t : term) (n : nat) (ℐ : vector) (ℐ_μ : table) (ℰ : list clos_i) (ℰ_μ : list (list clos_i)).

Definition stack_i := list clos_i.
Definition c_env_i := list clos_i.
Definition k_env_i := list stack_i.

Inductive state_i :=
| St_i (t : term) (n : nat) (ℐ : vector) (ℐ_μ : table) (ℰ : c_env_i) (ℰ_μ : k_env_i) (𝒮 : stack_i).

Notation "⟨ t , n , ℐ , ℐ_μ , ℰ , ℰ_μ , 𝒮 ⟩" := (St_i t n ℐ ℐ_μ ℰ ℰ_μ 𝒮) (at level 80).
Notation "[ t , n , ℐ , ℐ_μ , ℰ , ℰ_μ ]" := (Cl_i t n ℐ ℐ_μ ℰ ℰ_μ) (at level 80).
(*4.1.2 Evaluation rules *)
Reserved Notation "σ_1 ↝ '^i' σ_2" (at level 30).

Inductive step_i : state_i → state_i → Prop :=

| I_var : ∀ n n' l g t ℐ ℐ' ℐ_μ ℐ_μ' ℰ ℰ' ℰ_μ ℰ_μ' 𝒮,
(n  _-_  ℐ _(_ l _)_   _=_  g) → (ℰ _(g)_   _=_  [t,n', ℐ',ℐ_μ',ℰ',ℰ_μ'])/(⟨ ‘l‘ ,n, ℐ,ℐ_μ,ℰ,ℰ_μ,𝒮⟩ ↝^i⟨ t,n', ℐ',ℐ_μ',ℰ',ℰ_μ',𝒮⟩) 

| I_app : ∀ n t u ℐ ℐ_μ ℰ ℰ_μ 𝒮,
⟨(t _*_ u),n, ℐ,ℐ_μ,ℰ,ℰ_μ,𝒮⟩ ↝^i⟨ t,n, ℐ,ℐ_μ,ℰ,ℰ_μ,[u,n, ℐ,ℐ_μ,ℰ,ℰ_μ]::𝒮⟩ 

| I_abs : ∀ n t c ℐ ℐ_μ ℰ ℰ_μ S',
⟨ _λ_ t,n, ℐ,ℐ_μ,ℰ,ℰ_μ,c ::S'⟩ ↝^i⟨ t, (S n),(( S n) :: ℐ),ℐ_μ,c ::ℰ,ℰ_μ,S'⟩ 

| I_catch : ∀ n t ℐ ℐ_μ ℰ ℰ_μ 𝒮,
⟨get-context t,n, ℐ,ℐ_μ,ℰ,ℰ_μ,𝒮⟩ ↝^i⟨ t, n, ℐ,(ℐ ::ℐ_μ),ℰ,(𝒮 ::ℰ_μ),𝒮⟩ 

| I_throw : ∀ n t α ℐ ℐ' ℐ_μ ℰ ℰ_μ 𝒮 𝒮',
(ℐ_μ _(α)_   _=_  ℐ') → (ℰ_μ _(α)_   _=_  𝒮')/(⟨set-context α t,n, ℐ,ℐ_μ,ℰ,ℰ_μ,𝒮⟩ ↝^i⟨ t, n, ℐ',ℐ_μ,ℰ,ℰ_μ,𝒮'⟩) 

where "σ_1 ↝ '^i' σ_2" := (step_i σ_1 σ_2).

Lemma step_i_unique : ∀ σ_1 σ_2 σ_2', σ_1 ↝^i σ_2 → σ_1 ↝^i σ_2' → σ_2=σ_2'.
Proof.
  introv. revert_last.
  intros y1 H H0. revert H0. revert y1.
  induction H; 
    (intros y2 D2; inversion D2; repeat progress ( eauto || econstructor || gen_subst || crush )).
Qed.

Hint Resolve step_i_unique : gen_subst_db.
(*4.2 Lock-step simulation (-)^⋆ *)
Reserved Notation "c '^⋆=_c' c'" (at level 30).
Reserved Notation "𝒮 '^⋆=_s' 𝒮'" (at level 30).
Reserved Notation "ℰ '^⋆=_e' ℰ'" (at level 30).
Reserved Notation "ℰ_μ '^⋆=_k' ℰ_μ'" (at level 30).

Inductive translate_clos : clos_i → clos → Prop :=
| T_cl : ∀ n t u ℐ ℐ_μ ℰ ℰ' ℰ_μ ℰ_μ',
(↓_( n)^( ℐ,ℐ_μ)(t)=u) → (ℰ^⋆=_e ℰ') → (ℰ_μ^⋆=_k ℰ_μ')/([t,n, ℐ,ℐ_μ,ℰ,ℰ_μ]^⋆=_c [ u,ℰ',ℰ_μ']) 

where "c '^⋆=_c' c'" := (translate_clos c c')

with translate_stack : stack_i → stack → Prop :=
| T_stack_1 :
nil ^⋆=_s nil 

| T_stack_2 : ∀ c c' 𝒮 𝒮',
(c^⋆=_c c') → (𝒮^⋆=_s 𝒮')/((c ::𝒮)^⋆=_s (c' ::𝒮')) 

where "𝒮 '^⋆=_s' 𝒮'" := (translate_stack 𝒮 𝒮')

with translate_c_env : c_env_i → c_env → Prop :=
| T_c_env_1 :
nil ^⋆=_e nil 

| T_c_env_2 : ∀ c c' ℰ ℰ',
(c^⋆=_c c') → (ℰ^⋆=_e ℰ')/((c ::ℰ)^⋆=_e (c' ::ℰ')) 

where "ℰ '^⋆=_e' ℰ'" := (translate_c_env ℰ ℰ')

with translate_k_env : k_env_i → k_env → Prop :=
| T_k_env_1 :
nil^⋆=_k nil 

| T_k_env_2 : ∀ 𝒮 𝒮' ℰ_μ ℰ_μ',
(𝒮^⋆=_s 𝒮') → (ℰ_μ^⋆=_k ℰ_μ')/((𝒮 ::ℰ_μ)^⋆=_k (𝒮' ::ℰ_μ')) 

where "ℰ_μ '^⋆=_k' ℰ_μ'" := (translate_k_env ℰ_μ ℰ_μ').

Reserved Notation "σ '^⋆=_σ' σ'" (at level 30).

Inductive translate_st : state_i → state → Prop :=
| T_st : ∀ n t u ℐ ℐ_μ ℰ ℰ' ℰ_μ ℰ_μ' 𝒮 𝒮',
([ t,n, ℐ,ℐ_μ,ℰ,ℰ_μ]^⋆=_c [u,ℰ',ℰ_μ']) → (𝒮^⋆=_s 𝒮')/(⟨ t,n, ℐ,ℐ_μ,ℰ,ℰ_μ,𝒮⟩^⋆=_σ ⟨ u,ℰ',ℰ_μ',𝒮'⟩) 

where "σ '^⋆=_σ' σ'" := (translate_st σ σ').

Scheme translate_clos_mut := Induction for translate_clos Sort Prop
with translate_stack_mut := Induction for translate_stack Sort Prop
with translate_c_env_mut := Induction for translate_c_env Sort Prop
with translate_k_env_mut := Induction for translate_k_env Sort Prop.

Lemma translate_clos_unique : ∀ c c_2 c_2', c^⋆=_c c_2 → c^⋆=_c c_2' → c_2=c_2'
with translate_stack_unique : ∀ s s_2 s_2', s^⋆=_s s_2 → s^⋆=_s s_2' → s_2=s_2'
with translate_c_env_unique : ∀ e e_2 e_2', e^⋆=_e e_2 → e^⋆=_e e_2' → e_2=e_2'
with translate_k_env_unique : ∀ k k_2 k_2', k^⋆=_k k_2 → k^⋆=_k k_2' → k_2=k_2'.
Proof.
  Tactic Notation "apply_unique_tac" tactic(l) :=
    (unique_tac
       (l (unique_spec_1 translate_clos)
          (unique_spec_1 translate_stack)
          (unique_spec_1 translate_c_env)
          (unique_spec_1 translate_k_env))).
  - apply_unique_tac translate_clos_mut.
  - apply_unique_tac translate_stack_mut.
  - apply_unique_tac translate_c_env_mut.
  - apply_unique_tac translate_k_env_mut.
Qed.

Hint Resolve translate_clos_unique : gen_subst_db.
Hint Resolve translate_stack_unique : gen_subst_db.
Hint Resolve translate_c_env_unique : gen_subst_db.
Hint Resolve translate_k_env_unique : gen_subst_db.

Lemma translate_st_unique : ∀ σ σ_2 σ_2', σ^⋆=_σ σ_2 → σ^⋆=_σ σ_2' → σ_2=σ_2'.
Proof.
  simple_unique_tac.
Qed.

Hint Resolve translate_st_unique : gen_subst_db.
(*4.2.1 Soundness *)
Lemma fetch_sound :  ∀ ℰ ℰ', ℰ^⋆=_e ℰ'→∀ n c, ℰ _(n)_   _=_  c→∃c',c^⋆=_c c'∧ℰ' _(n)_   _=_  c'.
Proof.
  intros ℰ ℰ' D1 n c D2.
  revert ℰ' D1.
  induction D2.
  - intros ℰ' D1.
    inversion D1.
    exists c'.
    split.
    * assumption.
    * apply Fetch_1.
  - intros ℰ' D1.
    inversion D1.
    apply IHD2 in H3.
    destruct H3.
    destruct H3.
    exists x.
    split.    
    * exact H3.
    * apply Fetch_2.
      exact H4. 
Qed.

Lemma fetch_mu_sound :  ∀ n 𝒮 ℰ_μ, ℰ_μ _(n)_   _=_  𝒮→∀ℰ'_μ,ℰ_μ^⋆=_k ℰ'_μ→∃𝒮',𝒮^⋆=_s 𝒮'∧ℰ'_μ _(n)_   _=_  𝒮'.
Proof.
  intros n S ℰ_μ D2.
  induction D2.
  - intros ℰ_μ' D1.
    inversion D1.
    exists 𝒮'.
    { 
      split.
      - assumption.
      - apply Fetch_1.
    }
  - intros ℰ_μ' D1.
    inversion D1.
    apply IHD2 in H3.
    destruct H3.
    destruct H3.
    exists x.
    {
      split.    
      - exact H3.
      - apply Fetch_2.
        exact H4.
    }
Qed.

Theorem soundness :  ∀ σ_1 σ_2 σ_1', σ_1 ↝^i σ_2 → σ_1^⋆=_σ σ_1' → ∃σ_2', σ_1' ↝ σ_2' ∧ σ_2^⋆=_σ σ_2'.
Proof.
  intros σ_1 σ_2 σ_1' D1.
  induction D1.
  - intro D2.
    inversion D2.
    inversion H9.
    inversion H14.
    assert (H222 := (fetch_sound _ _ H21)).
    subst.
    apply H222 in H0.
    destruct H0.
    destruct x.
    exists (⟨t0, ℰ0, ℰ_μ0, 𝒮'⟩).
    { 
      split.
    - apply K_var.
      inversion H24.
      inversion H.
      subst.
      assert (H71 := fetch_unique _ _ _ _ _ H7 H1).
      subst.
      assert (H82 := minus_unique _ _ _ _ H8 H2).
      subst.
      destruct H0.
      assumption.
    - subst.
      apply T_st.
      destruct H0.
      assumption.       
      assumption.
    }
-  intro D2.
   inversion D2.
   inversion H7.
   inversion H12.
   subst.
   exists (⟨t', ℰ', ℰ_μ', [u', ℰ', ℰ_μ']::𝒮'⟩).
   {
     split.
   - apply K_app.
   - { 
        apply T_st.
      - apply T_cl.
        assumption.
        assumption.
        assumption.
      - apply T_stack_2.
        apply T_cl.
        assumption.
        assumption.
        assumption.
        assumption.
     }
   }
-  intro D2.
   inversion D2.
   inversion H7.
   inversion H12.
   inversion H8.
   subst.
   exists (⟨t', (c'::ℰ'), ℰ_μ', 𝒮'0⟩).
   split.
   * apply K_abs.
   * apply T_st.
      + apply T_cl.
         assumption.
         apply T_c_env_2.
         assumption.
         assumption.
         assumption.
     + assumption.
-  intro D2.
   inversion D2.
   inversion H7.
   inversion H12.
   subst.
   exists (⟨t', ℰ', (𝒮'::ℰ_μ'), 𝒮'⟩).
   split.
   * apply K_catch.
   *  apply T_st.
      + apply T_cl.
         assumption.
         assumption.
         apply T_k_env_2. 
         assumption.
         assumption.
     + assumption.
-  intro D2.
   inversion D2.
   inversion H9.
   inversion H14.
   subst.
   assert (H333 := (fetch_mu_sound _ _ _ H0 _ H22)).
   destruct H333 as [ 𝒮2 H444 ].
   destruct H444.
   exists (⟨t', ℰ', ℰ_μ', 𝒮2⟩).
   split.
   * apply K_throw.
      assumption.
   * apply T_st.
      + apply T_cl.
         assert (V2 := fetch_unique _ _ _ _ _ H25 H).
         subst.
         assumption.
         assumption.
         assumption.
      + assumption.
Qed.

(*4.2.2 Completeness *)
Lemma fetch_complete :  ∀ ℰ ℰ', ℰ^⋆=_e ℰ'→∀ n c', ℰ' _(n)_   _=_  c'→∃c,c^⋆=_c c'∧ℰ _(n)_   _=_  c.
Proof.
  intros ℰ ℰ' D1 n c D2.
  revert ℰ D1.
  induction D2.
  - intros ℰ D1.
    inversion D1.
    exists c.
    split.
    * assumption.
    * apply Fetch_1.
  - intros ℰ D1.
    inversion D1.
    apply IHD2 in H3.
    destruct H3.
    destruct H3.
    exists x.
    split.    
    * exact H3.
    * apply Fetch_2.
      exact H4.
Qed.

Lemma fetch_mu_complete :  ∀ n 𝒮' ℰ_μ', ℰ_μ' _(n)_   _=_  𝒮'→∀ℰ_μ,ℰ_μ^⋆=_k ℰ_μ'→∃𝒮,𝒮^⋆=_s 𝒮'∧ℰ_μ _(n)_   _=_  𝒮.
Proof.
  intros n S ℰ_μ' D2.
  induction D2.
  - intros ℰ_μ D1.
    inversion D1.
    exists 𝒮.
    { 
      split.
      - assumption.
      - apply Fetch_1.
    }
  - intros ℰ_μ D1.
    inversion D1.
    apply IHD2 in H3.
    destruct H3.
    destruct H3.
    exists x.
    {
      split.    
      - exact H3.
      - apply Fetch_2.
        exact H4.
    }
Qed.

Theorem completeness :  ∀ σ_1' σ_2' σ_1, σ_1' ↝ σ_2' → σ_1^⋆=_σ σ_1' → ∃σ_2, σ_1 ↝^i σ_2 ∧ σ_2^⋆=_σ σ_2'.
Proof.
  intros σ_1' σ_2' σ_1 D1.
  induction D1.
  - intro D2.
    inversion D2.
    inversion H3.
    inversion H10.
    assert (H222 := (fetch_complete _ _ H17)).
    subst.
    apply H222 in H.
    destruct H.
    destruct x.
    exists (⟨t0, n0, ℐ0, ℐ_μ0, ℰ1, ℰ_μ1, 𝒮0⟩).
    { 
      split.
    - eapply I_var.
      exact H24.
      inversion H.
      subst.
      exact H1.
    - subst.
      apply T_st.
      destruct H.
      assumption.       
      assumption.
    }
  - intro D2.
    inversion D2.
    inversion H2.
    inversion H9.
    subst.
    exists (⟨t2, n, ℐ, ℐ_μ, ℰ0, ℰ_μ0, [u2, n, ℐ, ℐ_μ, ℰ0, ℰ_μ0]::𝒮0⟩).
    {
      split.
      - apply I_app.
      - { 
          apply T_st.
          - apply T_cl.
            assumption.
            assumption.
            assumption.
          - apply T_stack_2.
            apply T_cl.
            assumption.
            assumption.
            assumption.
            assumption.
        }
    }
  - intro D2.
    inversion D2.
    inversion H2.
    inversion H9.
    inversion H5.
    subst.
    exists (⟨t2, (S n), (S n)::ℐ, ℐ_μ, (c0::ℰ0), ℰ_μ0, 𝒮1⟩).
    {
      split.
      - apply I_abs.
      - {
          apply T_st.
          - apply T_cl.
            assumption.
            apply T_c_env_2.
            assumption.
            assumption.
            assumption.
          - assumption.
        }
    } 
  - intro D2.
    inversion D2.
    inversion H2.
    inversion H9.
    subst.
    exists (⟨t2, n, ℐ, (ℐ::ℐ_μ), ℰ0, (𝒮0::ℰ_μ0), 𝒮0⟩).
    {
      split.
      - apply I_catch.
      - {
          apply T_st.
          - apply T_cl.
            assumption.
            assumption.
            apply T_k_env_2. 
            assumption.
            assumption.
          - assumption.
        }
    }
  - intro D2.
    inversion D2.
    inversion H3.
    inversion H10.
    subst.
    assert (H333 := (fetch_mu_complete _ _ _ H _ H18)).
    destruct H333 as [ 𝒮2 H444 ].
    destruct H444.
    exists (⟨t2, n, ℐ', ℐ_μ, ℰ0, ℰ_μ0, 𝒮2⟩).
    {
      split.
      - apply I_throw.
        assumption.
        assumption.
      - {
          apply T_st.
          - apply T_cl.
            subst.
            assumption.
            assumption.
            assumption.
          - assumption.
        }
    }
Qed.

(*4.3 Lock-step simulation (-)^⋄ *)
Reserved Notation "ℰ  _(_  n _-  k  _)_   _=_  c" (at level 30).

Inductive compute_l : c_env_i → nat → nat → clos_i → Prop :=

| Compute_l_1 : ∀ ℰ n g k c,
(n _-_ k  _=_  g) → (ℰ _(g)_  _=_  c)/(ℰ _(_ n  _- k _)_   _=_  c) 

where "ℰ  _(_  n _-  k  _)_   _=_  c" := (compute_l ℰ n k c).

Lemma compute_l_unique : ∀ ℰ n k c c', ℰ _(_ n  _- k _)_   _=_  c → ℰ _(_ n  _- k _)_   _=_  c' → c=c'.
Proof.
  introv. revert_last.
  intros y1 H H0. revert H0. revert y1.
  induction H; 
    (intros y2 D2; inversion D2; repeat progress ( eauto || econstructor || gen_subst || crush )).
Qed.

Hint Resolve compute_l_unique : gen_subst_db.

Reserved Notation "c '^⋄=_c' c'" (at level 30).
Reserved Notation "𝒮 '^⋄=_s' 𝒮'" (at level 30).
Reserved Notation "ℰ_μ '^⋄=_k' ℰ_μ'" (at level 30).

Inductive flatten : nat → c_env_i → vector → l_env_l →Prop :=
| T_flatten_1 : ∀ n ℰ,
flatten n ℰ nil  nil 

| T_flatten_2 : ∀ c c' n k ℰ ℐ ℒ,
(ℰ _(_ n  _- k _)_   _=_  c) → (c^⋄=_c c') → (flatten n ℰ ℐ ℒ)/(flatten n ℰ (k ::ℐ)  (c' ::ℒ)) 

with map_flatten : nat → c_env_i → table → m_env_l →Prop :=
| T_map_flatten_1 : ∀ n ℰ,
map_flatten n ℰ nil nil 

| T_map_flatten_2 : ∀ n ℰ ℐ ℒ ℐ_μ ℒ_μ,
(flatten n ℰ ℐ ℒ) → (map_flatten n ℰ ℐ_μ ℒ_μ)/(map_flatten n ℰ (ℐ ::ℐ_μ) (ℒ ::ℒ_μ)) 

with translate_clos_l : clos_i → clos_l → Prop :=
| T_cl_l : ∀ n t ℐ ℐ_μ ℒ ℒ_μ ℰ ℰ_μ ℰ_μ',
(flatten n ℰ ℐ ℒ) → (map_flatten n ℰ ℐ_μ ℒ_μ) → (ℰ_μ^⋄=_k ℰ_μ')/([t,n, ℐ,ℐ_μ,ℰ,ℰ_μ]^⋄=_c [ t,ℒ,ℒ_μ,ℰ_μ']) 

where "c '^⋄=_c' c'" := (translate_clos_l c c')

with translate_stack_l : stack_i → stack_l → Prop :=
| T_stack_l_1 :
nil ^⋄=_s nil 

| T_stack_l_2 : ∀ c c' 𝒮 𝒮',
(c^⋄=_c c') → (𝒮^⋄=_s 𝒮')/((c ::𝒮)^⋄=_s (c' ::𝒮')) 

where "𝒮 '^⋄=_s' 𝒮'" := (translate_stack_l 𝒮 𝒮')

with translate_k_env_l : k_env_i → k_env_l → Prop :=
| T_k_env_l_1 :
nil ^⋄=_k nil 

| T_k_env_l_2 : ∀ 𝒮 𝒮' ℰ_μ ℰ_μ',
(𝒮^⋄=_s 𝒮') → (ℰ_μ^⋄=_k ℰ_μ')/((𝒮 ::ℰ_μ)^⋄=_k (𝒮' ::ℰ_μ')) 

where "ℰ_μ '^⋄=_k' ℰ_μ'" := (translate_k_env_l ℰ_μ ℰ_μ').


Reserved Notation "σ '^⋄=_σ' σ'" (at level 30).

Inductive translate_st_l : state_i → state_l → Prop :=
| T_st_l : ∀ n t u ℐ ℐ_μ ℰ ℒ ℒ_μ ℰ_μ ℰ_μ' 𝒮 𝒮',
([ t,n, ℐ,ℐ_μ,ℰ,ℰ_μ]^⋄=_c [u,ℒ,ℒ_μ,ℰ_μ']) → (𝒮^⋄=_s 𝒮')/(⟨ t,n, ℐ,ℐ_μ,ℰ,ℰ_μ,𝒮⟩^⋄=_σ ⟨ u,ℒ,ℒ_μ,ℰ_μ',𝒮'⟩) 

where "σ '^⋄=_σ' σ'" := (translate_st_l σ σ').

Scheme flatten_mut := Induction for flatten Sort Prop
with map_flatten_mut := Induction for map_flatten Sort Prop
with translate_clos_l_mut := Induction for translate_clos_l Sort Prop
with translate_stack_l_mut := Induction for translate_stack_l Sort Prop
with translate_k_env_l_mut := Induction for translate_k_env_l Sort Prop.

Lemma flatten_unique : ∀ n ℰ ℐ ℒ ℒ', flatten n ℰ ℐ ℒ→flatten n ℰ ℐ ℒ'→ℒ=ℒ' 
with map_flatten_unique : ∀ n ℰ ℐ_μ ℒ_μ ℒ_μ', 
map_flatten n ℰ ℐ_μ ℒ_μ→map_flatten n ℰ ℐ_μ ℒ_μ'→ℒ_μ=ℒ_μ' 
with translate_clos_l_unique : ∀ c c_2 c_2', c^⋄=_c c_2 → c^⋄=_c c_2' → c_2=c_2'
with translate_stack_l_unique : ∀ s s_2 s_2', s^⋄=_s s_2 → s^⋄=_s s_2' → s_2=s_2'
with translate_k_env_l_unique : ∀ k k_2 k_2', k^⋄=_k k_2 → k^⋄=_k k_2' → k_2=k_2'.
Proof.
  Tactic Notation "translate_l_unique_tac" tactic(L) :=
    repeat clear_last;
    introv; revert_last;
    intros y1 H H0; revert H0; revert y1;
    apply (L
          (fun n ℰ ℐ ℒ h => ∀ ℒ', flatten n ℰ ℐ ℒ'→ℒ=ℒ')
          (fun n ℰ ℐ_μ ℒ_μ h => ∀ ℒ_μ', map_flatten n ℰ ℐ_μ ℒ_μ'→ℒ_μ=ℒ_μ')
          (fun c c_2 h => ∀ c_2', c^⋄=_c c_2' → c_2=c_2')
          (fun s s_2 h => ∀ s_2', s^⋄=_s s_2' → s_2=s_2')
          (fun k k_2 h => ∀ k_2', k^⋄=_k k_2' → k_2=k_2'));
    (intros; revert_last; intro D2; inversion D2;
     repeat progress ( eauto || econstructor || gen_subst || crush )).
  - translate_l_unique_tac flatten_mut.
  - translate_l_unique_tac map_flatten_mut.
  - translate_l_unique_tac translate_clos_l_mut.
  - translate_l_unique_tac translate_stack_l_mut.
  - translate_l_unique_tac translate_k_env_l_mut.
Qed.

Hint Resolve flatten_unique : gen_subst_db.
Hint Resolve map_flatten_unique : gen_subst_db.
Hint Resolve translate_clos_l_unique : gen_subst_db.
Hint Resolve translate_stack_l_unique : gen_subst_db.
Hint Resolve translate_k_env_l_unique : gen_subst_db.

Lemma translate_st_l_unique : ∀ σ σ_2 σ_2', σ^⋄=_σ σ_2 → σ^⋄=_σ σ_2' → σ_2=σ_2'.
Proof.
  introv. revert_last.
  intros y1 H H0. revert H0. revert y1.
  induction H; 
    (intros y2 D2; inversion D2; repeat progress ( eauto || econstructor || gen_subst || crush )).
Qed.

Hint Resolve translate_st_l_unique : gen_subst_db.
(*4.3.1 Soundness *)
Lemma minus_next : ∀ n k g, n _-_ k  _=_  g → S n _-_ k  _=_  S g.
Proof.
  intros n k g 𝒟_1.
  induction 𝒟_1.
  - apply Minus_1.
  - apply Minus_2.
    assumption.
Qed.

Lemma weaken_flatten : ∀ {n ℰ ℐ ℒ}, ∀ c', flatten n ℰ ℐ ℒ → flatten (S n) (c' ::ℰ) ℐ ℒ.
Proof.
  intros n ℰ ℐ ℒ c' 𝒟_1.
  induction 𝒟_1.
  - apply T_flatten_1.
  - inversion_clear H.
    assert (𝒟_4 := minus_next _ _ _ H1).
    {
      eapply T_flatten_2.
      - {
          eapply Compute_l_1.
          - exact 𝒟_4.
          - apply Fetch_2. 
            exact H2.
        }
      - assumption.
      - assumption.
    }
Qed.

Lemma weaken_map_flatten : ∀ {n ℰ ℐ_μ ℒ_μ}, ∀ c',
map_flatten n ℰ ℐ_μ ℒ_μ → map_flatten (S n) (c' ::ℰ) ℐ_μ ℒ_μ.
Proof.
  intros n ℰ ℐ_μ ℒ_μ c' 𝒟_1.
  induction 𝒟_1.
  - apply T_map_flatten_1.
  - eapply weaken_flatten in H.
    {
      apply  T_map_flatten_2.
      - exact H.
      - assumption.
    }
Qed.

Lemma map_flatten_sound : ∀ n ℰ ℐ' ℐ_μ α, 
ℐ_μ _(α)_  _=_  ℐ' → ∀ ℒ_μ, map_flatten n ℰ ℐ_μ ℒ_μ → ∃ ℒ', flatten n ℰ ℐ' ℒ' ∧  ℒ_μ _(α)_  _=_  ℒ'.
Proof.
  intros n ℰ ℐ' ℐ_μ α 𝒟_2.
  induction 𝒟_2.
  - intros ℒ_μ 𝒟_1.
    inversion 𝒟_1; subst.
    exists ℒ.
    split.
    exact H3.
    apply Fetch_1.
  - intros ℒ_μ 𝒟_1.
    inversion 𝒟_1; subst.
    assert (D34 := IH𝒟_2 _ H5).
    destruct D34.
    exists x.
    destruct H.
    split.
    exact H.
    apply Fetch_2.
    assumption.
Qed.

Lemma fetch_sound_l :  ∀ n l k c ℐ, 
ℐ _(l)_   _=_  k→∀ℰ,ℰ _(_ n  _- k _)_   _=_  c→∀ℒ,flatten n ℰ ℐ ℒ→∃c',c^⋄=_c c'∧ℒ _(l)_   _=_  c'.
Proof.
  intros n l k c ℐ D1.
  induction D1.
  - intros ℰ' D2 ℒ D3.
    inversion D3.
    inversion D2.
    inversion H1.
    subst.
    exists c'.
    assert (H666 := minus_unique _ _ _ _ H7 H13).
    subst. 
    assert (H555 := fetch_unique _ _ _ _ _ H8 H14).
    subst. 
    {
      split.
      - assumption.
      - apply Fetch_1.
    }
  - intros ℰ' D2 ℒ D3.
    inversion D3.
    subst.
    assert (H888 := IHD1 _ D2 _ H6).
    destruct H888.
    destruct H.
    exists x.
    {
      split.
      - assumption.
      - apply Fetch_2.
        assumption.
    }
Qed.

Lemma fetch_mu_sound_l :  ∀ n 𝒮 ℰ_μ, ℰ_μ _(n)_   _=_  𝒮→∀ℰ'_μ,ℰ_μ^⋄=_k ℰ'_μ→∃𝒮',𝒮^⋄=_s 𝒮'∧ℰ'_μ _(n)_   _=_  𝒮'.
Proof.
  intros n S ℰ_μ D2.
  induction D2.
  - intros ℰ_μ' D1.
    inversion D1.
    exists 𝒮'.
    { 
      split.
      - assumption.
      - apply Fetch_1.
    }
  - intros ℰ_μ' D1.
    inversion D1.
    apply IHD2 in H3.
    destruct H3.
    destruct H3.
    exists x.
    {
      split.    
      - exact H3.
      - apply Fetch_2.
        exact H4.
    }
Qed.

Theorem soundness_l :  ∀ σ_1 σ_2 σ_1', σ_1 ↝^i σ_2 → σ_1^⋄=_σ σ_1' → ∃σ_2', σ_1' ↝^l σ_2' ∧ σ_2^⋄=_σ σ_2'.
Proof.
  intros σ_1 σ_2 σ_1' D1.
  induction D1.
  - intro D2.
    inversion D2.
    inversion H9.
    inversion H.
    subst.
    assert (H111 := Compute_l_1 _ _ _ _ _ H25 H0).
    assert (H222 := fetch_sound_l _ _ _ _ _ H24 _ H111 _ H14).
    destruct H222.
    destruct H1.
    destruct x.
    exists (⟨t0, ℒ0, ℒ_μ0, ℰ_μ0, 𝒮'⟩).
    { 
      split.
      - apply L_var.
        assumption.
      - subst.
        apply T_st_l.
        destruct H0.
        assumption.       
        assumption.
        assumption.
    }
  - intro D2.
    inversion D2.
    inversion H7.
    subst.
    exists (⟨t, ℒ, ℒ_μ, ℰ_μ', [u, ℒ, ℒ_μ, ℰ_μ']::𝒮'⟩).
    {
      split.
      - apply L_app.
      - { 
          apply T_st_l.
          - apply T_cl_l.
            assumption.
            assumption.
            assumption.
          - apply T_stack_l_2.
            apply T_cl_l.
            assumption.
            assumption.
            assumption.
        assumption.
     }
   }
  - intro D2.
    inversion D2.
    inversion H7.
    inversion H8.
    subst.
    exists (⟨t, (c'::ℒ), ℒ_μ, ℰ_μ', 𝒮'0⟩).
    {
      split.
      - apply L_abs.
      - { 
          apply T_st_l.
          - { 
              apply T_cl_l.
              - {
                  eapply T_flatten_2.
                  - { 
                      eapply Compute_l_1.
                      - apply minus_id.
                      - apply Fetch_1.
                    }
                  - assumption.
                  - apply weaken_flatten.
                    assumption.
                }
              - apply weaken_map_flatten.
                assumption.
              - assumption.
            }
          - assumption.
        }
    }
  - intro D2.
    inversion D2.
    inversion H7.
    subst.
    exists (⟨t, ℒ, (ℒ::ℒ_μ), (𝒮'::ℰ_μ'), 𝒮'⟩).
    {   
      split.
      - apply L_catch.
      -  {
          apply T_st_l.
          - { 
              apply T_cl_l.
              - assumption.
              - { 
                  apply T_map_flatten_2.
                  - assumption.
                  - assumption.
                }
              - {
                  apply T_k_env_l_2. 
                  - assumption.
                  - assumption.
                }
            }
          - assumption.
        }
    }
  - intro D2.
    inversion D2.
    inversion H9.
    subst.
    assert (H333 := (fetch_mu_sound_l _ _ _ H0 _ H23)).
    destruct H333 as [ 𝒮2 H444 ].
    destruct H444.
    assert (H555 := (map_flatten_sound _ _ _ _ _ H _ H22)).
    destruct H555.
    destruct H3.
    exists (⟨t, x, ℒ_μ, ℰ_μ', 𝒮2⟩).
    { 
      split.
      - { 
          apply L_throw.
          - assumption.
          - assumption.
        }
      - {
          apply T_st_l.
          - { 
              apply T_cl_l.
              - assumption.
              - assumption.
              - assumption.
            }
          - assumption.
        }
    }
Qed.

(*4.3.2 Completeness *)
Lemma map_flatten_complete : ∀ {n ℰ ℒ' ℒ_μ α}, 
ℒ_μ _(α)_  _=_  ℒ' → ∀ ℐ_μ, map_flatten n ℰ ℐ_μ ℒ_μ → ∃ ℐ', flatten n ℰ ℐ' ℒ' ∧ ℐ_μ _(α)_  _=_  ℐ'.
Proof.
  intros n ℰ ℒ' ℒ_μ α 𝒟_2.
  induction 𝒟_2.
  - intros ℐ_μ 𝒟_1.
    inverts 𝒟_1 as D11 D12.
    exists ℐ0.
    {
      split.
      - exact D11.
      - apply Fetch_1.
    }
  - intros ℐ_μ 𝒟_1.
    inverts 𝒟_1 as D11 D12.
    assert (D34 := IH𝒟_2 _ D12).
    destruct D34 as [ℐ' D34].
    exists ℐ'.
    destruct D34 as [ D3 D4 ].
    {
      split.
      - exact D3.
      - apply Fetch_2.
        exact D4.
    }
Qed.

Lemma fetch_complete_l :  ∀ {n l c' ℐ ℰ ℒ}, 
flatten n ℰ ℐ ℒ→ℒ _(l)_   _=_  c'→∃ k c, ℰ _(_ n  _- k _)_   _=_  c ∧ c^⋄=_c c'∧ℐ _(l)_   _=_  k.
Proof.
  intros n l c' ℐ ℰ ℒ D1 D2.
  revert D1.
  revert ℐ.
  induction D2.
  - intros ℐ0 D1.
    inverts D1 as D31 D32 D33.
    exists k.
    exists c.
    { 
      split.
      - exact D31.
      - {
          split.
            - exact D32.
            - apply Fetch_1.
        }
    }
  - intros ℐ0 D1.
    inverts D1 as D31 D32 D33.
    assert (D2411 := IHD2 _ D33).
    destruct D2411 as [k0 D2411].
    destruct D2411 as [c0 D2411].
    exists k0 c0.
    destruct D2411 as [D20 D411].
    destruct D411 as [D4 D11].
    {
      split.    
      - exact D20.
      - {
          split.
          - exact D4.
          - apply Fetch_2.
            exact D11.
        }
    }
Qed.

Lemma fetch_mu_complete_l :  ∀ {n 𝒮' ℰ_μ'}, 
ℰ_μ' _(n)_   _=_  𝒮'→∀ℰ_μ,ℰ_μ^⋄=_k ℰ_μ'→∃𝒮,𝒮^⋄=_s 𝒮'∧ℰ_μ _(n)_   _=_  𝒮.
Proof.
  intros n S ℰ_μ' D2.
  induction D2.
  - intros ℰ_μ D1.
    inversion D1.
    exists 𝒮.
    { 
      split.
      - assumption.
      - apply Fetch_1.
    }
  - intros ℰ_μ D1.
    inversion D1.
    apply IHD2 in H3.
    destruct H3.
    destruct H3.
    exists x.
    {
      split.    
      - exact H3.
      - apply Fetch_2.
        exact H4.
    }
Qed.

Theorem completeness_l :  ∀ σ_1' σ_2' σ_1, σ_1' ↝^l σ_2' → σ_1^⋄=_σ σ_1' → ∃σ_2, σ_1 ↝^i σ_2 ∧ σ_2^⋄=_σ σ_2'.
Proof.
  intros σ_1' σ_2' σ_1 D1.
  induction D1.
  - intro D2.
    inverts keep D2 as D21 D22.
    inverts keep D21 as D211 D212 D213.
    assert (H222 := fetch_complete_l D211 H).
    destruct H222 as [ k0 H222 ].
    destruct H222 as [ c H222 ].
    destruct H222 as [ D112_12 D31_111 ].
    inverts D112_12 as D112 D12.
    destruct D31_111 as [ D31 D111 ].
    destruct c as [ t0 n0 ℐ0 ℐ_μ0 ℰ0 ℰ_μ1 ].
    exists (⟨t0, n0, ℐ0, ℐ_μ0, ℰ0, ℰ_μ1, 𝒮0 ⟩).
    { 
      split.
    - { 
        eapply I_var.
        - {
            eapply Compute_1.
            - exact D111.
            - exact D112.
          }
        - exact D12.
      }
    - {
        apply T_st_l.
        - exact D31.
        - exact D22.
      }
    }
  - intro D2.
    inverts D2 as D21 D22.
    inverts D21 as D211 D212 D213.
    exists (⟨t, n, ℐ, ℐ_μ, ℰ, ℰ_μ0, [u, n, ℐ, ℐ_μ, ℰ, ℰ_μ0]::𝒮0⟩).
    {
      split.
      - apply I_app.
      - { 
          apply T_st_l.
          - { 
              apply T_cl_l.
              - exact D211.
              - exact D212.
              - exact D213.
            }
          - { 
              apply T_stack_l_2.
              {
                apply T_cl_l.
                - exact D211.
                - exact D212.
                - exact D213.
              }
              - exact D22.
            }
        }
    }
  - intro D2.
    inverts D2 as D21 D22.
    inverts D21 as D211 D212 D213.
    inverts D22 as D221 D222.
    exists (⟨t, (S n), (S n)::ℐ, ℐ_μ, (c0::ℰ), ℰ_μ0, 𝒮0⟩).
    {
      split.
      - apply I_abs.
      - assert (D1 := minus_id (S n)). 
        assert (D2 := weaken_flatten c0 D211).
        assert (D3 := weaken_map_flatten c0 D212).
        { 
          apply T_st_l.
          - { 
              apply T_cl_l.
              - {
                  eapply T_flatten_2.
                  - {                  
                      eapply Compute_l_1.
                      - exact D1.
                      - apply Fetch_1.
                    }
                  - exact D221.
                  - exact D2.
                }
              - exact D3. 
              - exact D213.
            }
          - exact D222.
        }
    }
  - intro D2.
    inverts D2 as D21 D22.
    inverts D21 as D211 D212 D213.
    exists (⟨t, n, ℐ, (ℐ::ℐ_μ), ℰ, (𝒮0::ℰ_μ0), 𝒮0⟩).
    {
      split.
      - apply I_catch.
      - {
          apply T_st_l.
          - { 
              apply T_cl_l.
              - exact D211.
              - { 
                  eapply T_map_flatten_2.
                  -  exact D211.
                  -  exact D212.
                }
              - {
                  apply T_k_env_l_2.
                  - exact D22.
                  - exact D213.
                }
            }
          - exact D22.
        }
    }
  - intro D2.
    inverts D2 as D21 D22.
    inverts D21 as D211 D212 D213.
    assert (D32_12 := fetch_mu_complete_l H0 _ D213).
    assert (D5_11 := map_flatten_complete H _ D212).
    destruct D32_12 as [ 𝒮1 D32_12 ].
    destruct D32_12 as [ D32 D12 ].
    destruct D5_11 as [ ℐ1 D5_11 ].
    destruct D5_11 as [ D5 D11 ].
    exists (⟨t, n, ℐ1, ℐ_μ, ℰ, ℰ_μ0, 𝒮1⟩).
    {
      split.
      - {
          apply I_throw.
          - exact D11.
          - exact D12.
        }
      - {
          apply T_st_l.
          - { 
              apply T_cl_l.
              - exact D5.
              - exact D212.
              - exact D213.
            }
          - exact D32.
        }
    }
Qed.

(*
5 Conclusion and future work *)